//       (    )

//  
// () - 
// [] -   
// |  - 
// ?  -  
// .  -  
// +  -     
// *  -     
// {n} -    

#pragma warning(disable:4786)
#include<stdio.h>
#include<assert.h>
#include<stack>
#include<vector>
#include<algorithm>	// qsort
#include<set>		// debug
#include<string>	// debug

struct Node;
class List;
class Parser;
class DFA;


typedef unsigned char byte;
typedef unsigned int  uint;

inline void setmin(byte &sym1,byte sym2) {if(sym2<sym1) sym1=sym2;}
inline void setmax(byte &sym1,byte sym2) {if(sym2>sym1) sym1=sym2;}

namespace Debug {
	void PrintSym(byte,char [10]);
	void PrintSym(byte,byte,char [30]);
#ifdef _DEBUG
	//!   
	class Leak{
		static std::set<Leak*>set;
	public:
		Leak(const Leak&) {set.insert(this);}
		Leak() {set.insert(this);}
		~Leak() {set.erase(this);}
		static bool Check() {return !set.empty();}
	};
	std::set<Leak*>Leak::set;
#else
	class Leak{
	public:
	    static bool Check() {return false;}
	};
#endif
};

void Debug::PrintSym(byte sym,char buff[10])
{
	switch(sym) {
	case 0:    strcpy(buff,"min"); break;
	case 255:  strcpy(buff,"max"); break;
	default:   if(sym<33) sprintf(buff,"%d",(int)sym); else buff[0]=sym,buff[1]=0; break;
	}
}

//!      
void Debug::PrintSym(byte sym1,byte sym2,char buff[30]) 
{
	PrintSym(sym1,buff); buff=strchr(buff,0);
	if(sym1!=sym2) *buff++='-',PrintSym(sym2,buff);
}


//!     
class List:public Debug::Leak{
	uint hash;
	std::vector<Node*>list;
public:
	int Size()			{return list.size();}
	void Clear()		{list.clear();}
	void Add(Node *n)	{list.push_back(n);}
	bool isEmpty()		{return list.empty();}
	Node *At(int n)		{return list[n];}
	void Prepare();				//!<      
	int Compare(const List*);	//!<    
};


//!   ( )
struct Node:public Debug::Leak{
	// - Start                
	// - Sym              
	// - Con                 (       )
	// - Or               
	// - Removed             
	// - Open,Close        (  )
	// - Rep               
	// - State,StateAcc    
	enum e{Sym=1,Con,Or,Rep,ZeroOne,OneMore,ZeroMore,Accept,Open,Close,End,Removed,State,StateAcc,Start}op;
	int id;			//!<    ,  
	union{
		Node *p1;	//!<    
		Node *next;	//!<     (   Con  )
	};
	union{
		Node *p2;	//!<    
		Node *link; //!<    
	};
	union{
		struct{byte min,max;};
		List *list;	//!<    
	};
	static int last_id;

	Node(e p,Node *n1=0,Node *n2=0):op(p),p1(n1),p2(n2),min(0),max(0),id(0) {}
	Node(byte s1,byte s2,Node *n1=0,Node *n2=0):op(Sym),min(s1),max(s2),p1(n1),p2(n2),id(last_id++) {}
	Node(int c1,int c2,Node *n):op(Rep),min(c1),max(c2),p1(n),p2(0),id(0) {}
	Node(List *l):op(State),list(l),link(0),next(0),id(last_id++) {
		for(int n=0; n<l->Size(); n++) 
			if(list->At(n)->op==Accept) {op=StateAcc;break;}
	}
	//! 
	Node *Dup() {
		Node *n=new Node(*this);
		if(id) n->id=last_id++;
		return n;
	}
	//!  
	bool isIn(byte sym) {return sym>=min && sym<=max;}
	//!  
	void Free();
	//!       
	bool isLinked() {
		switch(op) {
		case ZeroOne:
		case OneMore:
		case ZeroMore:
		case Con:
		case Rep:
		case Sym: return true;
		}
		return !!link; //     .    ,      
	}
	//!    
	Node *Link(Node *n) {
		if(link) return new Node(Con,this,n);
		link=n; return this;
	}
	//!     
	void RemoveList() {
		assert(isState());
		delete list; list=0;
	}
	void Dump(const char*);
	bool isState() {return op==State || op==StateAcc;}
private:
	Node *Free(Node*);
	bool isSkipDump(std::set<Node*> &set);
	void DumpNode(FILE*,std::set<Node*>&,Node*);
	void DumpEdge(FILE*,std::set<Node*>&,Node*);
};
int Node::last_id=1;


Node *Node::Free(Node *list)
{
	if(!this || op==Removed) return list;
	if(op==State || op==StateAcc) RemoveList();
	op=Removed;
	list=p1->Free(list);
	list=p2->Free(list);
	next=list; 
	return this;
}

void Node::Free(){
	Node *list=Free(0);
	while(list) {
		Node *next=list->next;
		delete list;
		list=next;
	}
}


void List::Prepare()
{
	std::sort(list.begin(),list.end());	
	list.erase(std::unique(list.begin(),list.end()),list.end()); 
	hash=list.size();
	for(int n=0; n<list.size(); n++) hash=hash*137+list[n]->id;
}


int List::Compare(const List *obj)
{
	int s=hash-obj->hash;
	if(!s) {
		s=list.size()-obj->list.size();
		if(!s) for(int n=0; n<list.size(); n++) {
			s=list[n]-obj->list[n]; if(s) break;
		}
	}
	return s;
}


//!  
class Parser{
	//!    
	class Stack{
		std::stack<Node*>p; //!<  
	public:
		Node *Pop()			{Node *n=0; if(!p.empty()) n=p.top(),p.pop();return n;}
		void Push(Node *n)	{if(n) p.push(n);}
		void Reset()		{while(Pop());}
	}stack;
	void Scroll(Node::e);				//!<     
	Node *Dup(Node*);					//!<  
	const char *ParseSet(const char*);	//!<  
	const char *ParseRep(const char*);	//!<    
public:
	Node *BuildTree(const char*);		//!<      
	Node *BuildGraph(Node*,Node *f=new Node(Node::Accept)); //!<    
};

// ----- NFA -----

//! \param start  
//! \param end        
Node *Parser::BuildGraph(Node *start,Node *final)
{
	Node *t,*f,*del=start; 
	if(!start) return final;
	f=BuildGraph(start->p2,final);
	switch(start->op) {
	case Node::Con: 
		assert(start->link);
		start=BuildGraph(start->next,f);
		break;
	case Node::Sym:  
		start->link=f; del=0;
		break;
	case Node::ZeroOne:
		start->next=BuildGraph(start->next,f);
		start=new Node(Node::Or,start->next,f);
		break;
	case Node::OneMore:
		start=BuildGraph(start->next,t=new Node(Node::Or,f));
		t->p2=start;
		break;
	case Node::ZeroMore:
		start=new Node(Node::Or,start->next,f);
		start->next=BuildGraph(start->next,start);
		break;
	case Node::Rep:
		while(start->max-->1) f=BuildGraph(Dup(start->next),f);
		start=BuildGraph(start->next,f);
		break;
	case Node::Or:
		start->p1=BuildGraph(start->p1,final);
		start->p2=f; del=0;
		break;
	}
	delete del;
	return start;
}


void Parser::Scroll(Node::e op)
{
	Node *old,*last;
	for(;;) {
		last=stack.Pop();
		old =stack.Pop(); if(!old) break;
		if(old->isLinked() && last->isLinked()) stack.Push(old->Link(last));
		else if(old->op==Node::Or) {
			old->p1=stack.Pop();
			old->p2=last;
			stack.Push(old);
		} else if(op==Node::Close) {
			delete old;
			stack.Push(last);
			return;
		} else break;
	}
	stack.Push(old); stack.Push(last);
	if(op!=Node::End) stack.Push(new Node(op));
}


Node *Parser::Dup(Node *n)
{
	if(n) {
		n=n->Dup();
		n->p1=Dup(n->p1);
		n->p2=Dup(n->p2);
	}
	return n;
}


const char *Parser::ParseRep(const char *exp)
{
	int n=atoi(exp);
	stack.Push(new Node(n,n,stack.Pop()));
	while(*exp && *exp++!='}');
	return exp;
}


const char *Parser::ParseSet(const char *exp)
{
	byte sym1,sym2;
	Node *root=0;
	for(;;) {
		sym1=(byte)*exp++; assert(sym1);
		if(sym1==']') break;
		if(exp[0]=='-') sym2=exp[1],exp+=2; else sym2=sym1;
		Node *node=new Node(sym1,sym2);
		if(root) root=new Node(Node::Or,root,node);	else root=node;
	}
	stack.Push(root);
	return exp;
}


Node *Parser::BuildTree(const char *exp)
{
	stack.Reset();
	for(;;) {
		byte sym=(byte)*exp++; if(!sym) break;
		switch(sym) {
		case '|': Scroll(Node::Or); break;
		case '(': stack.Push(new Node(Node::Open)); break;
		case ')': Scroll(Node::Close); break;
		case '+': stack.Push(new Node(Node::OneMore ,stack.Pop())); break;
		case '?': stack.Push(new Node(Node::ZeroOne ,stack.Pop())); break;
		case '*': stack.Push(new Node(Node::ZeroMore,stack.Pop())); break;
		case '[': exp=ParseSet(exp); break;
		case '{': exp=ParseRep(exp); break;
		case '.': stack.Push(new Node(' ',(byte)255));  break;
		default:  stack.Push(new Node(sym,sym)); break;
		}
	}
	Scroll(Node::End);
	return stack.Pop();
}



// ----- DFA -----

//!     
class DFA{
	Node *graph;	//!<  
	Node *curr;		//!<   
	Node *start;	//!<    (    )
	List work_list;	//!<    
	Node *Shift(Node*,byte);	//!<      
	Node *AddSet();				//!<       
	void Add(Node*);			//!<      
	void Add(Node *node,byte sym,byte &min,byte &max);	//!<         
public:
	DFA():graph(0),curr(0),start(0) {}
	~DFA();
	void Init(Node*);	//!<     
	void Start();		//!<   
	int  Next(char);	//!<      ( )
	void FullBuild();	//!<       
	void Dump(const char *name) {start->Dump(name);}
};


DFA::~DFA()
{
	start->Free();
}

void DFA::Start()
{
	curr=0;
}

void DFA::Init(Node *g)
{
	graph=g;
	Node::last_id=1;
	work_list.Clear();
	Add(graph);
	AddSet();
}



//! 1    (   id );
//! -1     ;
//! 0    
int DFA::Next(char sym)
{
	if(!curr ) curr=start;
	curr=Shift(curr,(byte)sym);
	if(curr->op==Node::StateAcc) return 1;
	if(curr==start) return -1;
	return 0;
}


void DFA::Add(Node *node)
{
	if(!node) return;
	if(node->op==Node::Or) Add(node->p1),Add(node->p2); else work_list.Add(node);
}

void DFA::Add(Node *node,byte sym,byte &min,byte &max)
{
	assert(!node->isState());
	if(!node) return;		//          
	if(node->isIn(sym)) {	//     
		setmax(min,node->min);
		setmin(max,node->max);
		Add(node->link);
	} else {	//          
		assert(node->op==Node::Sym || node->op==Node::Accept || node->op==Node::Or);
		if(sym<node->min)      setmin(max,node->max-1);
		else if(sym>node->max) setmax(min,node->min+1);
	}
}


Node *DFA::AddSet()
{
	work_list.Prepare();
	Node *node,**ref=&start;
	while((node=*ref)) {
		assert(node->isState());
		if(!work_list.Compare(node->list)) return node;
		ref=&node->next;
	}
	node=*ref=new Node(new List(work_list));
	return node;
}


Node *DFA::Shift(Node *from,byte sym)
{
	assert(from->isState());
	//   
	for(Node *t=from->link;t;t=t->next) if(t->isIn(sym)) return t->link;
	//     
	if(!from->list) return start;
	byte min=0,max=255; 
	work_list.Clear();
	for(uint n=0; n<from->list->Size(); n++) Add(from->list->At(n),sym,min,max);
	if(work_list.isEmpty()) return start;
	Add(graph);
	Node *to=AddSet(); //     
	from->link=new Node(min,max,from->link,to);
	return to;
}


void DFA::FullBuild()
{
	Node *it; int n;
	for(it=start; it; it=it->next) {
		assert(it->isState());
		List *list=it->list;
		for(n=0; n<list->Size(); n++) {
			curr=it,Next(list->At(n)->min);
			curr=it,Next(list->At(n)->max);
		}
	}
#ifndef _DEBUG
	for(it=start; it; it=it->next) it->RemoveList();
#endif
}


// ----- RegExp -----

class RegExp{
	DFA dfa;		//!<  
	Node *graph;	//!<   
public:
	RegExp():graph(0) {}
	~RegExp() {graph->Free();}
	void SetExp(const char*);		//!<    
	void Compile();					//!<    ()
	const char *Find(const char*);	//!<    
};

void RegExp::SetExp(const char *exp)
{
	Parser obj; 
	graph->Free();
	graph=obj.BuildTree(exp);    graph->Dump("tree.dot");	//  
	graph=obj.BuildGraph(graph); graph->Dump("graph.dot");	//  
	dfa.Init(graph);
}

void RegExp::Compile()
{
	dfa.FullBuild();
	dfa.Dump("full_dfa.dot");	// DFA      
}

const char *RegExp::Find(const char *text)
{
	if(!graph) return 0;
	const char *res=0;
	int old=-1,id;
	do {
		id=dfa.Next(*text);
		if(old<0 && id>=0) res=text;
		if(id>0) break;
		old=id;
	}while(*text++);
	if(id<=0) res=0;
	dfa.Dump("dfa.dot");	// DFA     
	return res;
}


// ----- debug -----

bool Node::isSkipDump(std::set<Node*> &set)
{	
	if(!this) return true;
	if(set.find(this)!=set.end()) return true; 
	set.insert(this);
	return false;
}


//  ()  
void Node::DumpNode(FILE *file,std::set<Node*> &set,Node *parent)
{
	const char *shape=0; 
	if(this) {
		//if(op==Node::Or) {
		//	p1->DumpNode(file,set,parent);
		//	p2->DumpNode(file,set,parent);
		//	return;
		//}
		if(parent) fprintf(file,"  %d -> %d\n",parent,this);
	}
	if(isSkipDump(set)) return;
	std::string label; char buff[50]; 
	int n,id=this->id; 
	switch(op) {
	case Node::Rep:      
		sprintf(buff,"{%d,%d}",(int)min,(int)max); 
		label=buff; 
		break;
	case Node::Sym:      
		Debug::PrintSym(min,max,buff); 
		label=buff; 
		if(!shape) shape=""; 
		break;
	case Node::Con:      label="Con"; break;
	case Node::Accept:   label="Accept"; shape=",fontcolor=green"; break;
	case Node::Or:       label="Or"; break;
	case Node::Start:    label+="Start"; break; 
	case Node::OneMore:  label+='+'; break; 
	case Node::ZeroOne:  label+='?'; break;
	case Node::ZeroMore: label+='*'; break;
	case Node::StateAcc: shape=",fontcolor=green";
	case Node::State: 	 
		if(list) {
			id=0; 
			for(n=0; n<list->Size(); n++) 
				if(list->At(n)->op==Node::Accept) label+="[acc]";
				else sprintf(buff,"[%d]",list->At(n)->id),label+=buff;
		}
		break;
	default: label="<err>"; break;
	}
	if(!shape) shape=",fontcolor=gray";
	if(id) sprintf(buff,"[%d]\\n",id); else buff[0]=0;
	fprintf(file,"  %d [label=\"%s%s\"%s]\n",this,buff,label.c_str(),shape);
	if(isState()) {
		link->DumpEdge(file,set,this);
		next->DumpNode(file,set,0);
	}else{
		p1->DumpNode(file,set,this);
		p2->DumpNode(file,set,this);
	}
}


void Node::DumpEdge(FILE *file,std::set<Node*> &set,Node *parent)
{
	if(isSkipDump(set)) return;
	assert(!isState());
	char label[40];
	Debug::PrintSym(min,max,label);
	if(parent) fprintf(file,"  %d -> %d [label=\"%s\"]\n",parent,link,label);
	next->DumpEdge(file,set,parent);
}


void Node::Dump(const char *filename)
{
	std::set<Node*>set;
	FILE *file=fopen(filename,"wt");
	if(!file) return;
	fprintf(file,
		"# www.graphviz.org\n"
		"digraph G{\n"
		"  node [shape=none]\n"
		"  rankdir=LR\n"
		);
	Node (Node::Start,this).DumpNode(file,set,0);
	fprintf(file,"}\n");
	fclose(file);
}



int main()
{
	{
		RegExp re;
	//	re.SetExp("x?y+z*");
	//	re.SetExp("c{1}(a{2}|b){2}");
	//	re.SetExp("c(aa|b)(aa|b)");
	//	re.SetExp("fixed|prefix");
		re.SetExp("[a-z]+@([a-z]+[.])+[a-z]+");
		re.Compile(); //  (  )
		const char *t=re.Find("---a@a.baayaabxabaabababaaaxbabaababaxbaaaabababa  www.cca.com/ zcxcv");
		printf("%s\n",t?t:"<not found>");
	}
	if(Debug::Leak::Check()) printf("memory leak\n\a");
	system("pause");
	return 0;
}

